@@ -18,32 +18,42 @@ import calendar
 import logging
 from oslo_config import cfg
 from oslo_config import types
-import oslo_messaging
 from oslo_utils import encodeutils
 from oslo_utils import timeutils
+import re
 import six
 
 from searchlight.common import exception
 import searchlight.elasticsearch
+from searchlight.elasticsearch.plugins import helper
 from searchlight.elasticsearch.plugins import utils
 from searchlight.elasticsearch import ROLE_USER_FIELD
-from searchlight import i18n
+from searchlight.i18n import _
 from searchlight import plugin
 
-LOG = logging.getLogger(__name__)
-_LW = i18n._LW
-_LI = i18n._LI
-_ = i18n._
 
+LOG = logging.getLogger(__name__)
+
+resource_group_reg = re.compile(r'^[a-z0-9][a-z0-9_]*$')
 indexer_opts = [
-    cfg.StrOpt('resource_group_name', default="searchlight",
-               help="The default base name for accessing Elasticsearch"),
-    cfg.StrOpt('notifications_topic', default="searchlight_indexer",
+    cfg.Opt('resource_group_name', default="searchlight",
+            help="The default base name for accessing Elasticsearch",
+            type=types.String(regex=resource_group_reg)),
+    cfg.StrOpt('notifications_topic', default="notifications",
                help="The default messaging notifications topic"),
     cfg.BoolOpt('mapping_use_doc_values', default=True,
                 help='Use doc_values for mapped fields where applicable.'
                      'Allows lower memory usage at the cost of some disk '
-                     'space. Recommended, especially in large deployments.')
+                     'space. Recommended, especially in large deployments.'),
+    cfg.BoolOpt('include_region_name', default=False,
+                help='Whether or not to include region_name as a mapping '
+                     'field and in documents. The value will be determined '
+                     'from service_credentials and override_region_name for '
+                     'each plugin.'),
+    cfg.ListOpt('publishers',
+                help='The global publishers configuration, '
+                     'plugin can override this by setting their own publishers'
+                )
 ]
 
 CONF = cfg.CONF
@@ -67,7 +77,7 @@ class IndexBase(plugin.Plugin):
     @property
     def index_helper(self):
         if not getattr(self, '_index_helper', None):
-            self._index_helper = utils.IndexingHelper(self)
+            self._index_helper = helper.IndexingHelper(self)
         return self._index_helper
 
     @property
@@ -85,12 +95,26 @@ class IndexBase(plugin.Plugin):
     @property
     def resource_group_name(self):
         if not getattr(self, '_group_name', None):
-            # We no longer allow this to be set per-plugin. See note in
-            # get_plugin_opts()
-            if cfg.CONF.resource_plugin.resource_group_name is not None:
-                self._group_name = cfg.CONF.resource_plugin.resource_group_name
+            if self.parent_plugin:
+                self._group_name = self.parent_plugin.resource_group_name
+                if self.options.resource_group_name is not None and  \
+                        self.options.resource_group_name != self._group_name:
+                    LOG.warning(
+                        "Overriding resource_group for %(plugin)s because it "
+                        "differs from parent plugin %(parent)s resource_group "
+                        "%(resource_group)s" %
+                        {"plugin": self.document_type,
+                         "parent": self.parent_plugin.document_type,
+                         "resource_group": self._group_name}
+                    )
             else:
-                self._group_name = "searchlight"
+                if self.options.resource_group_name is not None:
+                    self._group_name = self.options.resource_group_name
+                elif cfg.CONF.resource_plugin.resource_group_name is not None:
+                    self._group_name = \
+                        cfg.CONF.resource_plugin.resource_group_name
+                else:
+                    self._group_name = "searchlight"
         return self._group_name
 
     @property
@@ -101,16 +125,29 @@ class IndexBase(plugin.Plugin):
     def alias_name_search(self):
         return "%s-search" % self.resource_group_name
 
+    @abc.abstractproperty
+    def resource_allowed_policy_target(self):
+        """Should return the policy target that this plugin's related service
+        uses to determine who has API access to list resources. Can return None
+        if this service doesn't support policy.
+        """
+        return None
+
+    @abc.abstractproperty
+    def service_type(self):
+        """Should match the keystone catalog service type for a resource"""
+        return None
+
     def prepare_index(self, index_name):
         """Prepare a new index for usage with this listener. We need to be
-           caled immediately after a new index is created, but before it
+           called immediately after a new index is created, but before it
            gets associated with an alias. Prepping means we will add
            the settings and mapping for this listener's document type.
         """
         if self.parent_plugin_type():
-            LOG.debug(_LI(
-                "Skipping index prep for %(doc_type)s; will be handled by"
-                "parent (%(parent_type)s)") %
+            LOG.debug(
+                "Skipping index prep for %(doc_type)s; will be handled by "
+                "parent (%(parent_type)s)" %
                 {"doc_type": self.document_type,
                  "parent_type": self.parent_plugin_type()})
             return
@@ -120,26 +157,8 @@ class IndexBase(plugin.Plugin):
             child_plugin.check_mapping_sort_fields()
 
         # Prepare the new index for this document type.
-        self.setup_index_settings(index_name=index_name)
         self.setup_index_mapping(index_name=index_name)
 
-    def initial_indexing(self, index_name=None, setup_data=True):
-        """Add data for this resource type. This method is called per plugin.
-           The assumption is that the aliases/indexes have already been setup
-           correctly before calling us. See the comments in the method
-           cmd/manage.py::sync() for more details.
-        """
-        if self.parent_plugin_type():
-            LOG.debug(_LI(
-                "Skipping initialization for %(doc_type)s; will be handled by"
-                "parent (%(parent_type)s)") %
-                {"doc_type": self.document_type,
-                 "parent_type": self.parent_plugin_type()})
-            return
-
-        if setup_data:
-            self.setup_data(index_name)
-
     def setup_index_settings(self, index_name):
         """Update index settings. """
         index_settings = self.get_settings()
@@ -156,12 +175,11 @@ class IndexBase(plugin.Plugin):
                                             doc_type=doc_type,
                                             body=mapping)
 
-    def setup_data(self, index=None):
-        """Insert all objects from database into search engine.
-           We are assuming this helper method is called by initial_indexing().
-           If you wish to call this method from somewhere else, please make
-           sure you understand the usage of the "index" parameter. See the
-           comment in plugins/utils.py::save_documents() for more details.
+    def index_initial_data(self, index_name=None):
+        """Insert all objects from a plugin (generally by API requests to its
+        service) into an index assumed to have been created with
+        prepare_index. If index_name is not set, the searchlight-listener alias
+        will be used instead.
         """
         object_list = self.get_objects()
         documents = []
@@ -169,18 +187,20 @@ class IndexBase(plugin.Plugin):
         for obj in object_list:
             document = self.serialize(obj)
             documents.append(document)
-            version = self.NotificationHandlerCls.get_version(document)
+            if self.NotificationHandlerCls:
+                version = self.NotificationHandlerCls.get_version(document)
+            else:
+                version = NotificationBase.get_version(document)
             versions.append(version)
         self.index_helper.save_documents(documents, versions=versions,
-                                         index=index)
+                                         index=index_name)
 
-        for child_plugin in self.child_plugins:
-            child_plugin.setup_data(index)
-
-    def get_facets(self, request_context, all_projects=False, limit_terms=0):
+    def get_facets(self, request_context, all_projects=False, limit_terms=0,
+                   include_fields=True, exclude_options=False):
         """Get facets available for searching, in the form of a list of
         dicts with keys "name", "type" and optionally "options" if a field
-        should have discreet allowed values
+        should have discreet allowed values. If include_fields is false,
+        only the total doc count will be requested.
         """
         exclude_facets = self.facets_excluded
         is_admin = request_context.is_admin
@@ -194,36 +214,89 @@ class IndexBase(plugin.Plugin):
 
             return False
 
-        def get_facets_for(mapping, prefix=''):
-            facets = []
-            for name, properties in six.iteritems(mapping):
-                if properties.get('type') == 'nested':
+        def get_facets_for(property_mapping, meta_mapping, prefix='',
+                           inside_nested=None):
+            mapping_facets = []
+            for name, properties in property_mapping.items():
+                property_type = properties.get('type', 'object')
+                if property_type in ('nested', 'object'):
                     if include_facet(prefix + name):
-                        facets.extend(get_facets_for(properties['properties'],
-                                                     "%s%s." % (prefix, name)))
+                        is_nested = property_type == 'nested'
+                        mapping_facets.extend(
+                            get_facets_for(properties['properties'],
+                                           meta_mapping,
+                                           "%s%s." % (prefix, name),
+                                           inside_nested=is_nested))
                 else:
                     indexed = properties.get('index', None) != 'no'
                     if indexed and include_facet(name):
-                        facets.append({
-                            'name': prefix + name,
-                            'type': properties['type']
-                        })
-            return facets
+                        facet_name = prefix + name
+                        mapping_facet = {
+                            'name': facet_name,
+                            'type': property_type
+                        }
 
-        facets = get_facets_for(self.get_mapping()['properties'])
+                        # If we're inside either an object or nested object,
+                        # add that to the mapping too
+                        if inside_nested is not None:
+                            mapping_facet['nested'] = inside_nested
+
+                        # Add raw field if it's an analyzed field,
+                        # aggregation on analyzed string field doesn't
+                        # work well
+                        if (properties['type'] == 'string' and
+                                properties.get('index') != 'not_analyzed' and
+                                'raw' in properties.get('fields', {})):
+                            mapping_facet['facet_field'] = facet_name + '.raw'
+
+                        # Plugin can specify _meta mapping to link an id with a
+                        # resource type, which can be used by Client program/UI
+                        # to GET more information from the resource type.
+                        # See https://www.elastic.co/guide/en/elasticsearch/
+                        # reference/2.1/mapping-meta-field.html
+                        if facet_name in meta_mapping:
+                            mapping_facet.update(meta_mapping[facet_name])
+
+                        if (self.get_parent_id_field() and
+                                name == self.get_parent_id_field()):
+                            mapping_facet['parent'] = True
+                            if 'resource_type' not in mapping_facet:
+                                mapping_facet['resource_type'] = \
+                                    self.parent_plugin_type()
+
+                        mapping_facets.append(mapping_facet)
+
+            return mapping_facets
+
+        facets = []
+        if include_fields:
+            facets = get_facets_for(self.get_mapping()['properties'],
+                                    self.get_mapping().get('_meta', {}))
 
         # Don't retrieve facet terms for any excluded fields
         included_fields = set(f['name'] for f in facets)
-        facet_terms_for = set(self.facets_with_options) & included_fields
-        facet_terms = self._get_facet_terms(facet_terms_for,
-                                            request_context,
-                                            all_projects,
-                                            limit_terms)
-        for facet in facets:
-            if facet['name'] in facet_terms:
-                facet['options'] = facet_terms[facet['name']]
+        options_fields = set(self.facets_with_options) & included_fields
 
-        return facets
+        raw_fields = dict([(f['name'], f['facet_field'])
+                           for f in facets
+                           if f.get('facet_field')])
+        # If field has a raw field, use a tuple of field name and
+        # raw field name
+        facet_terms_for = [(field, raw_fields[field])
+                           if field in raw_fields
+                           else field
+                           for field in options_fields]
+
+        facet_terms, doc_count = self._get_facet_terms(
+            facet_terms_for, request_context, all_projects,
+            limit_terms=limit_terms, exclude_options=exclude_options)
+
+        if include_fields:
+            for facet in facets:
+                if facet['name'] in facet_terms:
+                    facet['options'] = facet_terms[facet['name']]
+
+        return facets, doc_count
 
     @property
     def facets_excluded(self):
@@ -238,53 +311,85 @@ class IndexBase(plugin.Plugin):
         return ()
 
     def _get_facet_terms(self, fields, request_context,
-                         all_projects, limit_terms):
-        term_aggregations = utils.get_facets_query(fields, limit_terms)
+                         all_projects, limit_terms,
+                         exclude_options=False):
+        # Fields can be empty if there are no facet terms desired,
+        # but we will run a size=0 search to get the doc count. If the
+        # user does not want the options associated with the facet fields,
+        # do not aggregate. This is controlled with a parameter to the
+        # API call.
+        body = {}
+        term_aggregations = {}
 
-        if term_aggregations:
-            body = {
-                'aggs': term_aggregations,
+        if fields:
+            mapping = self.get_mapping()['properties']
+            # Nested fields will be all the mapped fields with type=='nested'
+            # Not tremendously keen on the way this is structured but the
+            # nested fields are a bit special case
+            nested_fields = [name
+                             for name, properties in mapping.items()
+                             if properties['type'] == 'nested']
+            if not exclude_options:
+                term_aggregations = utils.get_facets_query(fields,
+                                                           nested_fields,
+                                                           limit_terms)
+            if term_aggregations:
+                body['aggs'] = term_aggregations
+
+        role_filter = request_context.user_role_filter
+
+        filter_query = {
+            "bool": {
+                "filter": {
+                    "bool": {
+                        "must": {
+                            "term": {ROLE_USER_FIELD: role_filter}
+                        }
+                    }
+                }
             }
+        }
 
-            role_filter = request_context.user_role_filter
-            plugin_filters = [{
-                "term": {ROLE_USER_FIELD: role_filter}
-            }]
-            if not (request_context.is_admin and all_projects):
-                plugin_filters.extend(
-                    self._get_rbac_field_filters(request_context))
+        # Add in the RBAC filters unless all_projects is requested
+        if not (request_context.is_admin and all_projects):
+            rbac_filters = self._get_rbac_field_filters(request_context)
 
-            body['query'] = {
-                "filtered": {
-                    "filter": {
-                        "and": plugin_filters
-                    }}}
+            # minimum_should_match:1 is assumed in filter context,
+            # but I'm including it explicitly so nobody spends an hour
+            # scouring the documentation to check that is the case
+            if rbac_filters:
+                filter_query["bool"]["filter"]["bool"].update(
+                    {"should": self._get_rbac_field_filters(request_context),
+                     "minimum_should_match": 1})
 
-            results = self.engine.search(
-                index=self.alias_name_search,
-                doc_type=self.get_document_type(),
-                body=body,
-                ignore_unavailable=True,
-                search_type='count')
+        body['query'] = filter_query
 
-            agg_results = results.get('aggregations', {})
-            facet_terms = utils.transform_facets_results(
-                agg_results,
-                self.get_document_type())
+        results = self.engine.search(
+            index=self.alias_name_search,
+            doc_type=self.get_document_type(),
+            body=body,
+            ignore_unavailable=True,
+            size=0)
 
-            if not agg_results:
-                LOG.warning(_LW(
-                    "No aggregations found for %(resource_type)s. There may "
-                    "be a mapping problem.") %
-                    {'resource_type': self.get_document_type()})
-            return facet_terms
-        return {}
+        agg_results = results.get('aggregations', {})
+        doc_count = results['hits']['total']
+
+        facet_terms = utils.transform_facets_results(
+            agg_results,
+            self.get_document_type())
+
+        if term_aggregations and not agg_results:
+            LOG.warning(
+                "No aggregations found for %(resource_type)s. There may "
+                "be a mapping problem." %
+                {'resource_type': self.get_document_type()})
+        return facet_terms, doc_count
 
     def check_mapping_sort_fields(self):
-        """Check that fields that are expected to define a 'raw' field so so"""
+        """Check that fields that are expected to define a 'raw' field do so"""
         fields_needing_raw = searchlight.elasticsearch.RAW_SORT_FIELDS
         mapped_properties = self.get_mapping().get('properties', {})
-        for field_name, field_mapping in six.iteritems(mapped_properties):
+        for field_name, field_mapping in mapped_properties.items():
             if field_name in fields_needing_raw:
                 raw = field_mapping.get('fields', {}).get('raw', None)
                 if not raw:
@@ -328,16 +433,45 @@ class IndexBase(plugin.Plugin):
 
     @property
     def routing_field(self):
-        """Whatever field should be treated as the routing value.
-        This is required for plugins which want to base all the CRUD
-        operations based on the _routing definition.
+        """Whatever field should be treated as the routing value. Whatever is
+        used must be available to all notifications on the child since
+        index/update/delete operations will all require it. Additionally, if
+        it is NOT the parent id, the parent plugin must be routed using the
+        same field. This field is accessed AFTER serialization.
         """
         return None
 
+    @property
+    def include_region_name(self):
+        return cfg.CONF.resource_plugin.include_region_name
+
+    @property
+    def region_name(self):
+        """Returns a region name if enabled by config and configured in
+        service_credentials. Adding it as per-plugin rather than global
+        in case at some point per-plugin credentials are supported.
+        """
+        # Check the override in each plugin's options first, then use
+        # whatever's in service_credentials
+        region_name = self.options.override_region_name
+        if region_name:
+            return region_name
+        region_name = getattr(CONF.service_credentials,
+                              'os_region_name', None)
+        if region_name:
+            return [region_name]
+
     @property
     def enabled(self):
         return self.options.enabled
 
+    @property
+    def publishers(self):
+        if self.options.publishers is not None:
+            return self.options.publishers
+        elif cfg.CONF.resource_plugin.publishers is not None:
+            return cfg.CONF.resource_plugin.publishers
+
     @property
     def allow_admin_ignore_rbac(self):
         """If set for a plugin, an administrative query for all_projects will
@@ -351,22 +485,13 @@ class IndexBase(plugin.Plugin):
 
         This is in the format of OS::Service::Resource typically.
         """
-        raise NotImplemented()
+        raise NotImplementedError()
 
     def register_parent(self, parent):
         if not self.parent_plugin:
             parent.child_plugins.append(self)
             self.parent_plugin = parent
 
-    def get_index_display_name(self, indent_level=0):
-        """The string used to list this plugin when indexing"""
-        display = '\n' + '    ' * indent_level + '--> ' if indent_level else ''
-
-        display += '%s' % (self.document_type)
-        display += ''.join(c.get_index_display_name(indent_level + 1)
-                           for c in self.child_plugins)
-        return display
-
     def get_query_filters(self, request_context, ignore_rbac=False):
         """Gets an index/type filter as es json filter dsl, and include
         plugin-specific RBAC filter terms unless `ignore_rbac` is True
@@ -375,27 +500,39 @@ class IndexBase(plugin.Plugin):
         query_filter = {
             'indices': {
                 'index': self.alias_name_search,
-                'no_match_filter': 'none',
-                'filter': {
-                    'and': [
-                        {'type': {'value': self.get_document_type()}}
-                    ]
+                'no_match_query': 'none',
+                'query': {
+                    'bool': {
+                        'filter': {
+                            'bool': {
+                                'must': {
+                                    'type': {
+                                        'value': self.get_document_type()
+                                    }
+                                }
+                            }
+                        }
+                    }
                 }
             }
         }
 
         if not (ignore_rbac and self.allow_admin_ignore_rbac):
             rbac_filters = self._get_rbac_field_filters(request_context)
-            query_filter['indices']['filter']['and'].extend(rbac_filters)
+            if rbac_filters:
+                bool_clause = query_filter['indices']['query']['bool']
+                bool_clause['filter']['bool'].update({
+                    'should': rbac_filters,
+                    'minimum_should_match': 1})
 
         return query_filter
 
     @abc.abstractmethod
     def _get_rbac_field_filters(self, request_context):
-        """Return any RBAC field filters to be injected into an indices
-        query. Document type will be added to this list.
+        """Return any RBAC field filters in a list to be injected into an
+        indices query. Document type will be added. The filters will be used
+        in the 'should' filter clause (i.e. OR).
         """
-        return []
 
     def get_notification_handler(self):
         """Get the notification handler which implements NotificationBase."""
@@ -405,22 +542,17 @@ class IndexBase(plugin.Plugin):
         return None
 
     def filter_result(self, hit, request_context):
-        """Filter each outgoing search result; document in hit['_source']. By
-        default, this does nothing since information shouldn't be indexed.
+        """Filter each outgoing search result; document in hit['_source'].
         """
-        pass
-
-    def get_settings(self):
-        """Get an index settings."""
-        return {
-            "index": {
-                "gc_deletes": CONF.elasticsearch.index_gc_deletes
-            }
-        }
+        if 'highlight' in hit:
+            # We need to clean up any fields 'leaked' by highlighting.
+            # As a reminder since this is a rare case, the 'highlight'
+            # dict is a peer of '_source' in the overall hit structure.
+            hit['highlight'].pop(ROLE_USER_FIELD, None)
 
+    @abc.abstractmethod
     def get_mapping(self):
         """Get an index mapping."""
-        return {}
 
     def get_full_mapping(self):
         """Gets the full mapping doc for this type, including children. This
@@ -433,17 +565,26 @@ class IndexBase(plugin.Plugin):
 
         type_mapping = self.get_mapping()
 
-        def apply_rbac_field(mapping):
-            mapping['properties'][ROLE_USER_FIELD] = {
+        # Add common mapping fields
+        # ROLE_USER_FIELD is required for RBAC by all plugins
+        type_mapping['properties'][ROLE_USER_FIELD] = {
+            'type': 'string',
+            'index': 'not_analyzed',
+            'include_in_all': False
+        }
+
+        # Add region name
+        if self.include_region_name:
+            type_mapping['properties']['region_name'] = {
                 'type': 'string',
                 'index': 'not_analyzed',
-                'include_in_all': False
             }
 
-        apply_rbac_field(type_mapping)
+        if 'updated_at' not in type_mapping['properties'].keys():
+            type_mapping['properties']['updated_at'] = {'type': 'date'}
 
         if self.mapping_use_doc_values:
-            utils.IndexingHelper.apply_doc_values(type_mapping)
+            helper.IndexingHelper.apply_doc_values(type_mapping)
 
         expected_parent_type = self.parent_plugin_type()
         mapping_parent = type_mapping.get('_parent', None)
@@ -491,15 +632,23 @@ class IndexBase(plugin.Plugin):
 
     @classmethod
     def get_plugin_opts(cls):
-        """Options that can be overridden per plugin. Note that
-        resource_group_name is not present while we determine a fix for
-        running has_parent/has_child queries across indices with different
-        sets of types (https://bugs.launchpad.net/searchlight/+bug/1558240)
+        """Options that can be overridden per plugin.
         """
         opts = [
+            cfg.StrOpt("resource_group_name"),
             cfg.BoolOpt("enabled", default=cls.is_plugin_enabled_by_default()),
             cfg.StrOpt("admin_only_fields"),
-            cfg.BoolOpt('mapping_use_doc_values')
+            cfg.BoolOpt('mapping_use_doc_values'),
+            cfg.ListOpt('override_region_name',
+                        help="Override the region name configured in "
+                             "'service_credentials'. This is useful when a "
+                             "service is deployed as a cloud-wide service "
+                             "rather than per region (e.g. Region1,Region2)."),
+            cfg.ListOpt('publishers',
+                        help='Used to configure publishers for the plugin, '
+                             'value could be publisher names configured in '
+                             'setup.cfg file.'
+                        )
         ]
         if cls.NotificationHandlerCls:
             opts.extend(cls.NotificationHandlerCls.get_plugin_opts())
@@ -525,6 +674,15 @@ class NotificationBase(object):
         """Get the list of event types this plugin responds to."""
         return list(six.iterkeys(self.get_event_handlers()))
 
+    def get_log_fields(self, event_type, payload):
+        """Return an iterable of key value pairs in payload that will be
+        present in notification log messages. Document type, event type, timing
+        and tenant information will be prepended.
+        """
+        if 'id' in payload:
+            return ('id', payload['id']),
+        return ()
+
     @abc.abstractmethod
     def get_event_handlers(self):
         """Returns a mapping of event name to function"""
@@ -550,13 +708,15 @@ class NotificationBase(object):
 
     def process(self, ctxt, publisher_id, event_type, payload, metadata):
         """Process the incoming notification message."""
-        LOG.debug("Received %s event for %s",
-                  event_type,
-                  self.index_helper.document_type)
         try:
-            self.get_event_handlers()[event_type](payload,
-                                                  metadata['timestamp'])
-            return oslo_messaging.NotificationResult.HANDLED
+            docs = self.get_event_handlers()[event_type](
+                event_type,
+                payload,
+                metadata['timestamp'])
+            if docs:
+                if not hasattr(docs, '__iter__'):
+                    docs = [docs]
+            return docs
         except Exception as e:
             LOG.error(encodeutils.exception_to_unicode(e))
 
@@ -604,7 +764,7 @@ class NotificationBase(object):
         We can address this before that happens; worst case is that you'd have
         to reindex everything, or reset the version.
 
-        The "timestamp" has similiar issues. When the "timestamp" overflowed
+        The "timestamp" has similar issues. When the "timestamp" overflowed
         the 9-digit field, time becomes indistinguishable. The 9 digits
         millisecond precision gives us around 27 hours. It should be enough to
         distinguish notifications with different timestamps.
@@ -612,7 +772,7 @@ class NotificationBase(object):
         updated = None
 
         # Pick the best/preferred date field to calculate version from
-        date_fields = ('updated_at', 'created_at')
+        date_fields = ('updated_at', 'created_at', 'deleted_at')
         if preferred_date_field:
             date_fields = (preferred_date_field,) + date_fields
 
@@ -622,8 +782,9 @@ class NotificationBase(object):
                 break
         else:
             date_fields_str = ', '.join(date_fields)
-            msg = ('Failed to build elasticsearch version; none of %s'
-                   'found in payload: %s' % (date_fields_str, payload))
+            msg = ('Failed to build elasticsearch version; none of %(dfs)s '
+                   'found in payload: %(payload)s' %
+                   {'dfs': date_fields_str, 'payload': payload})
             raise exception.SearchlightException(message=msg)
 
         updated_obj = timeutils.parse_isotime(updated)
